"""
Parser
======

This contains methods and classes to manage parsing of code
"""

from copy import deepcopy
from collections import namedtuple

from pygments.lexers import PythonConsoleLexer, PythonLexer, BashSessionLexer
from pygments.token import String, Token

# =============================================================================
# Pygments Token Management
# =============================================================================

class PurdyLexer:
    _registry = {
        'con':('con', 'Python 3 Console', PythonConsoleLexer, True, 'code'), 
        'py3':('py3', 'Python 3 Source', PythonLexer, False, 'code'),
        'bash':('bash', 'Bash Console', BashSessionLexer, True, 'code'),
    }

    def __init__(self, name, description, pygments_lexer_cls, is_console,
            palette):
        self.name = name
        self.description = description
        self.pygments_lexer = pygments_lexer_cls()
        self.is_console = is_console
        self.palette = palette

    @classmethod
    def factory_from_source(cls, source):
        if source.startswith('>>> ') or '\n>>> ' in source:
            return PurdyLexer(*cls._registry['con'])
        elif source.startswith('$ ') or '\n$ ' in source:
            return PurdyLexer(*cls._registry['bash'])

        return PurdyLexer(*cls._registry['py3'])

    @classmethod
    def factory_from_name(cls, name):
        return PurdyLexer(*cls._registry[name])

    @classmethod
    def choices(cls):
        output = []
        for key, value in cls._registry.items():
            output.append(f'"{key}" ({value[1]})')

        return ', '.join([f'"{key}" ({value[1]})' for key, value in \
            cls._registry.items()])

    @classmethod
    def names(cls):
        return cls._registry.keys()

# -----------------------------------------------------------------------------

def token_is_a(token1, token2):
    """Returns true if token1 is the same type as or a child type of token2"""
    if token1 == token2:
        return True

    parent = token1.parent
    while(parent != None):
        if parent == token2:
            return True

        parent = parent.parent

    return False


def token_ancestor(token, ancestor_list):
    """Tokens are hierarchical, in some situations you need to translate a
    token into one from a known list, e.g. turning a
    "Token.Literal.Number.Integer" into a "Number". This method takes a token
    and a list of approved ancestors and attempts to make the map. If no
    ancestor is found then a generic "Token" object is returned

    :param token: token to translate into an approved ancestor
    :param ancestor_list: list of approved ancestor tokens
    """
    if token in ancestor_list:
        return token

    # token not in the approved list, search its ancestors
    token = token.parent
    while(token != None):
        if token in ancestor_list:
            return token

        token = token.parent

    # something went wrong with our lookup, return the default
    return Token

# ===========================================================================
# Purdy Code Representation 
# ===========================================================================

CodePart = namedtuple('CodePart', ['token', 'text'])


class BlankCodeLine:
    def render_line(self, colourizer):
        return ''


class FoldedCodeLine:
    def __init__(self, size):
        self.size = size

    def __eq__(self, other):
        if self.size == other.size:
            return True

        return False

    def render_line(self, colourizer):
        return colourizer.colourize(self)


class CodeLine:
    def __init__(self, parts, lexer, line_number=-1, highlight=False):
        """Represents a displayed line of code.

        :param parts: list of :code:`CodePart` objects that correspond to 
                       this line of code
        :param lexer: PurdyLexer used to parse the content
        :param line_number: line number for the line, -1 for off (default)
        :param highlight: True if this line is currently highlighted. Defaults
                          to False.
        """
        # too many bugs caused by a change to the parts list after the
        # CodeLine was created, copy the damn thing so it is internal only
        self.parts = deepcopy(parts)       
        self.lexer = lexer
        self.line_number = line_number
        self.highlight = highlight

        self.text = ''.join([part.text for part in parts])

    def __str__(self):
        num = ''
        if self.line_number > -1:
            num = f'{self.line_number:3} '
        return f'CodeLine("{num}{self.text}")'

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        # for testing purposes we want a comparison of content values, not of
        # the references
        if self.line_number != other.line_number or \
                self.highlight != other.highlight or \
                self.lexer.__class__ != other.lexer.__class__:
            return False

        for x, part, in enumerate(self.parts):
            if part.token != other.parts[x].token or \
                    part.text != other.parts[x].text:
                return False

        return True

    def render_line(self, colourizer):
        return colourizer.colourize(self)


def parse_source(source, lexer):
    """Parses blocks of source text, returning a list of :class:`CodeLine` 
    objects.
    """
    parser = _Parser(lexer)
    parser.parse(source)

    return parser.lines


class _Parser:
    def __init__(self, lexer):
        self.lexer = lexer
        self.lines = []

    def parse(self, content):
        self.parts = []
        for token_type, text in self.lexer.pygments_lexer.get_tokens(content):
            if text.startswith('\n'):
                self.newline_handler(token_type)
                if len(text) > 1:
                    # something came after the \n, handle it
                    self.string_handler(token_type, text[1:])
            elif text == '':
                # tokenizer sometimes puts in empty stuff, skip it
                continue
            elif token_is_a(token_type, String) and '\n' in text:
                self.string_handler(token_type, text)
            else:
                self.default_handler(token_type, text)

    def newline_handler(self, token):
        # hit a CR, time to create a new CodeLine object
        if not self.parts:
            part = CodePart(token, '')
            self.parts = [part, ]

        self.lines.append( CodeLine(self.parts, self.lexer) )

        # reset to start the next set of tokens
        self.parts = []

    def string_handler(self, token, text):
        # String tokens may be multi-line
        for line in text.splitlines(True):
            part = CodePart(token, line.rstrip('\n'))
            self.parts.append(part)

            if line[-1] == '\n':
                self.lines.append( CodeLine(self.parts, self.lexer) )
                self.parts = []

    def default_handler(self, token, text):
        if text[-1] == '\n':
            # there is a \n at the end of the text, need to rebuild it
            # without it, then create the CodeLine object
            part = CodePart(token, text.rstrip('\n'))
            self.parts.append(part)

            # text caused a CR, create a new CodeLine object
            self.lines.append( CodeLine(self.parts, self.lexer ) )

            # reset to start the next group of tokens
            self.parts = []
        else:
            part = CodePart(token, text)
            self.parts.append(part)



from pygments.token import Keyword, Name, Comment, String, Error, \
    Number, Operator, Generic, Token, Whitespace, Punctuation, Text

from purdy.parser import FoldedCodeLine, token_ancestor

# =============================================================================
# Urwid Colourizer

_code_palette = {
    # urwid colour spec supports both 16 and 256 colour terminals
    #                    fg16            bg16    fg256   bg256
    Token:              ('',             '', '', '',             ''),
    Whitespace:         ('',             '', '', '',             ''),
    Comment:            ('dark cyan',    '', '', '#6dd',         ''),
    Keyword:            ('brown',        '', '', '#d8d',         ''),
    Operator:           ('brown',        '', '', '#aaa',         ''),
    Punctuation:        ('dark cyan',    '', '', '#8df',         ''),
    Text:               ('dark cyan',    '', '', '#ddd',         ''),
    Name:               ('light gray',   '', '', '#ddd',         ''), 
    Name.Builtin:       ('dark cyan',    '', '', '#8af',    ''), 
    Name.Builtin.Pseudo:('dark cyan',    '', '', '#a66,bold',    ''), 
    Name.Function:      ('dark cyan',    '', '', '#adf',   ''),
    Name.Class:         ('dark cyan',    '', '', '#adf',   ''),
    Name.Exception:     ('dark green',   '', '', '#fd6,bold',   ''),
    Name.Decorator:     ('dark cyan',    '', '', '#fd6,bold',    ''),
    String:             ('dark magenta', '', '', '#ddd', ''),
    Number:             ('dark magenta', '', '', '#f86', ''),
    Generic.Prompt:     ('dark blue',    '', '', '#aaa',    ''),
    Generic.Error:      ('dark green',   '', '', '#fd6,bold',   ''),
    Generic.Traceback:  ('',             '', '', '#ddd',    ''),
    Error:              ('dark green',   '', '', '#fd6,bold',         ''),
}

_xml_palette = dict(_code_palette)
_xml_palette.update({
    Name.Attribute: ('brown',        '', '', 'brown',        ''),
    Keyword:        ('dark cyan',    '', '', '#8af',         ''),
    Name.Tag:       ('dark cyan',    '', '', '#8af',         ''),
    Punctuation:    ('dark cyan',    '', '', '#8af',         ''),
})

class UrwidColourizer:
    palettes = {
        'code':_code_palette,
        'xml':_xml_palette,
    }

    @classmethod
    def create_urwid_palette(cls):
        """Returns a list of colour tuples that Urwid uses as its palette. The
        list is based on the UrwidColourizer.colours with a couple extra items
        """
        urwid_palette = []
        for name, palette in cls.palettes.items():

            for key, value in palette.items():
                # for each item in our colours hash create a tuple consisting of
                # the token name and its values
                item = (f'{name}_{key}', ) + value
                urwid_palette.append( item )

                # do it again for highlighted tokens, for 16 colour mode change
                # both the fg and bg colour, for 256 colour mode just change the
                # background
                item = (f'{name}_{key}_highlight', 'black', 'light gray', '', 
                    value[3], 'g23')
                urwid_palette.append( item )

        # add miscellaneous other palette items
        urwid_palette.extend([
            ('reverse', 'black', 'white', '', 'black', 'white'),
            ('bold', 'white,bold', '', '', 'white,bold', ''),
            ('title', 'white,underline', '', '', 'white,underline', ''),
            ('folded', 'white', '', '', 'white', ''),
            ('line_number', 'dark gray', '', '', 'dark gray', ''),
            ('empty', '', '', '', '', ''),
            ('empty_highlight', '', 'light gray', '', '', 'g23'),
        ])
        
        return urwid_palette

    @classmethod
    def colourize(cls, code_line):
        """Returns a list containing markup tuples as used by urwid.Text
        widgets.

        :param code_line: a :class:`CodeLine`   object to colourize
        """
        if isinstance(code_line, FoldedCodeLine):
            return ('folded', '     ⋮')

        palette = code_line.lexer.palette
        ancestor_list = cls.palettes[palette].keys()

        output = []
        if code_line.line_number >= 0:
            output.append( cls.line_number(code_line.line_number) )

        for part in code_line.parts:
            ancestor = token_ancestor(part.token, ancestor_list)
            key = f'{palette}_{ancestor}' 

            if code_line.highlight:
                key += '_highlight'

            # Urwid uses a palette which has been built as a hash using the
            # names of the ancestor tokens as keys and the fg/bg colour
            # choices as values, each piece of marked up text is a tuple of
            # the palette key and the text to display
            markup = (key, part.text)
            output.append(markup)

        return output

    @classmethod
    def line_number(cls, num):
        """Returns a colourized version of a line number"""
        return ('line_number', f'{num:3} ')
